Requirements: You've mastered the basics of Dockers through 1.1 and 1.2

We now pivot from consistent Python interpreter and packages to consistent NodeJS packages.

Like usual, our approach is to work locally first, then upload the Dockerfile online to test if the same output is produced.

Let's work locally first.

**1. Create a NodeJS script at local development**

app.js:
```
const express = require('express');
const app = express();

console.log('Hello, my NodeJS version is expected to be around 18, and it actually is:', process.version);
console.log('And my Express version is expected to be around 4.21.2, and it actually is:', require('express/package.json').version);
```

**2. Prepare for Dockerfile**

Create package.json:
```
{
  "dependencies": {
    "express": "4.21.2"
  }
}
```

Create a .dockerignore:
```
.DS_Store
node_modules
package-lock.json
__pycache__
*.pyc
.env
```

**3. Write a Dockerfile**

The Dockerfile is where you define the configuration of your Docker image. Here's an example:
```
# Use the specific NodeJS base image (bundled in Debian)
FROM node:18

# Set the working directory in the container.
# All below commands will work from working directory.
WORKDIR /app

# Copy the package.json file to the container
COPY package.json .

# Install the NodeJS packages
RUN npm install

# Copy your application files (if any)
COPY . .

# Specify the command to run your application (optional, depends on your use case)
CMD ["node", "app.js"]
```

^ Although `COPY . .` is sufficient if we had moved that line to above the `RUN npm install` command, having `COPY package.json` is just more explicit and communicates the intention to other developers.

**4. Build the Docker Image**
Run the following command in the directory where your `Dockerfile` is located:

```
docker build -t consistent_node
```

This will create a Docker image with your exact NodeJS version.

> [!note] Default tag to "latest"
> Since we are not specifying a version this time, by default, Docker assigns the tag "latest".
> So equivalently:  `docker build -t consistent_node:latest .`
> 
> Running `docker images` will show:
> REPOSITORY                       TAG        IMAGE ID       CREATED          SIZE
> consistent_node                  latest     1c1491610513   4 minutes ago    1.1GB

**5. Run the Docker Container**

Once the image is built, you can create and run a container from it:
```
docker run -it consistent_node
```

Your output could be:
```
Hello, my NodeJS version is expected to be around 18, and it actually is: v18.20.7

And my Express version is expected to be around 4.21.2, and it actually is: 4.21.2
```

This runs once then returns to your shell session because app.js doesn't keep the script process continually running like a NodeJS Express does.

> [!note] Cat OS
> Since you can run a virtual container off the image, here you can prove that the NodeJS base image includes Debian. Run:
> `docker run --rm consistent_node cat /etc/os-release`
> The `--rm` is just a best practice to close the container after running the command in the container's shell and getting the output, though not needed here since the app.js does not keep opened.

**6. Test on remote system:**
Copy your folder of files (with the app.js, package.json, Dockerfile, .dockerignore) into a remote server. Then SSH and repeat the steps to check it's the same output when you build then run the image.